! $Id: FixedFormForEverImplementation.f90 2794 2019-03-21 10:28:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the FixedFormForEverImplementation module.


!*******************************************************************************
!!
!> Procedures to implement the FixedFormForEver program.

MODULE FixedFormForEverImplementation
  
  USE StNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: Execute
  
  !> Format specification to use for reporting general errors, that may 
  !! not be specific to a particular file.
  CHARACTER(*), PARAMETER :: fmt_err = "(A)"
  
  !-----------------------------------------------------------------------------
  
  !> Number of continuous lines for a source file - the number 
  !! of continuation lines plus one for the initial line.
  INTEGER, PARAMETER :: block_length = 256
  
  !-----------------------------------------------------------------------------
  ! Fixed form stuff.
  
  !> Number of valid positions for a statement in the statement field of a 
  !! fixed form source file.
  INTEGER, PARAMETER :: statement_field_length = 72 - 6
  
  !> Type that encapsulates the "output as fixed form" action.
  TYPE, EXTENDS(StNodeQueryVisitor) :: fixed_formifier_visitor
    !> Logical unit for output of results.
    INTEGER :: unit
    !> Character buffer for storing current chunk.
    !!
    !! The length is the maximum number of characters for a contiguous 
    !! section of fixed form source, given by the statement field length 
    !! multipled by the maximum number of continuation lines plus one, 
    !! to account for the initial line.
    CHARACTER(statement_field_length * block_length) :: buffer = ''
    !> Current position in the character buffer.
    !!
    !! This indicates the last character placed in the buffer.  Consequently 
    !! never greater than LEN(buffer).
    INTEGER :: pos = 0
    !> Statement label for the initial line.
    !!
    !! Zero means no label.
    INTEGER :: label = 0
    !> Bad flag - an excessively long statement (for fixed form) was 
    !! encountered.
    LOGICAL :: bad = .FALSE.
  CONTAINS
    PROCEDURE :: ExecuteNode => ffv_ExecuteNode
    PROCEDURE :: ExecuteStmt => ffv_ExecuteStmt
  END TYPE fixed_formifier_visitor
  
  ! Some configuration options to work around marginal cases and compiler bugs.
  !
  ! All are worded such that I consider .FALSE. or zero conforming.
  
  !> Force any statement that can be used to end a program unit to be 
  !! on its own line, even if that statement is not ending a program unit.
  !!
  !! This practically is a superset of isg_program_unit_end_never_continued, 
  !! because (in the absence of an excessively long statement) if the 
  !! statement is the only thing on a line it will not be continued.
  LOGICAL, PARAMETER :: isg_program_unit_end_never_followed = .FALSE.
  
  !> Prevent continuation of any statement that can be used to end a 
  !! program unit, even if that statement is not ending a program 
  !! unit.
  LOGICAL, PARAMETER :: isg_program_unit_end_never_continued = .FALSE.
  
  !> Consider context when determining whether the initial line of 
  !! a statement appears to be a program unit END statement.
  !!
  !! This is subverted in part by isg_program_unit_end_never_continued.
  !!
  !! - 0: Do not consider any context.  Any initial line that looks like 
  !!      END, END PROGRAM, END BLOCKDATA, etc, will be considered to 
  !!      appear as a program unit end statement, and not be emitted.
  !! - 1: Consider context.  The nature of the current program unit will 
  !!      be considered when determining whether a statement appears to 
  !!      be a program unit end statement (i.e. ENDBLOCKDATA is not 
  !!      a valid program end statement in a main program).  The name 
  !!      of the program unit will not be considered.
  !! - 2: Consider context, including the name of the program unit 
  !!      being compiled.
  INTEGER, PARAMETER :: consider_context_for_appears_like_end = 0
  
  !-----------------------------------------------------------------------------
  ! Free form stuff.
  
  !> Number of characters available per free form line.
  INTEGER, PARAMETER :: line_length = 132
  
  !> Type that encapsulates the compact free form operation.
  TYPE, EXTENDS(StNodeQueryVisitor) :: free_formifier_visitor
    !> Logical unit for output of results.
    INTEGER :: unit
    !> Character buffer for storing current chunk.
    CHARACTER(line_length) :: buffer(block_length) = ''
    !> Current column in the character buffer.
    !!
    !! This indicates the last character placed in the buffer.  Consequently 
    !! never greater than LEN(buffer), may be zero when no characters are 
    !! yet to be written to the line.
    INTEGER :: column = 0
    !> Current line in the character buffer.  This always ranges between 
    !! one and SIZE(buffer) inclusive.
    INTEGER :: line = 1
    !> List of lines that were excessively long.
    INTEGER, ALLOCATABLE :: bad_lines(:)
  CONTAINS
    PROCEDURE :: ExecuteNode => freev_ExecuteNode
    PROCEDURE :: ExecuteStmt => freev_ExecuteStmt
  END TYPE free_formifier_visitor
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Fixed formify some fortran source.
  !!
  !! @param[in]     file_list         List of the file specifications that 
  !! nominate the files to include in the fixed-formication.  The mapping 
  !! from file specifications to file names is defined by the host.
  !!
  !! @param[in]     console_read      'FREE' to read input source as 
  !! free form from the console, 'FIXED' to read input source as fixed 
  !! form from the console, '' to not read from the console.
  !!
  !! @param[in]     free              Try out this new fangled free form 
  !! thing.
  !!
  !! @param[in]     out_unit          Logical unit connected for formatted 
  !! output to use for reporting the results of the fixed-formification.
  !!
  !! @param[in]     err_unit          Logical unit connected for formatted 
  !! output to use for reporting any errors.
  !!
  !! @param[in]     stop_code         Error code - non-zero on error with 
  !! a value consistent with the stop codes reported by the main program.
  !!
  !! I keep writing fornication.  This subroutine doesn't do that.
  !!
  !! For each file we build the statement tree, merging the trees in to 
  !! a single tree as we go.  Failure to build the tree for any file 
  !! terminates execution.  A tree traversal visitor operation then 
  !! renders the statements in the combined tree in the requested form.
  
  SUBROUTINE Execute( file_list, console_read, free,  &
      out_unit, err_unit, stop_code )
    
    USE BaseStNodes
    USE BuildTreePass
    USE CompilerHosts
    USE DefaultHosts
    USE Scanner
    USE Sources
    USE SourceUnits
    USE Strings
    USE Errors
    
    USE ScanFixedForm
    USE ScanFreeForm
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: INPUT_UNIT
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(String), INTENT(IN) :: file_list(:)
    CHARACTER(*), INTENT(IN) :: console_read
    LOGICAL, INTENT(IN) :: free
    INTEGER, INTENT(IN) :: out_unit
    INTEGER, INTENT(IN) :: err_unit
    INTEGER, INTENT(OUT) :: stop_code
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! File names
    TYPE(String), ALLOCATABLE :: filenames(:)
    
    ! Error object for procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    ! Compiler host object to define host behaviour.
    TYPE(DefaultHost) :: host
    
    INTEGER :: ie             ! Error index.
    INTEGER :: if             ! File index.
    
    ! The combined statement trees.
    TYPE(BaseStNode), POINTER :: combined
    
    ! The statement tree for the current file.
    TYPE(BaseStNode), POINTER :: current
    
    ! The object the encapsulates the fixedformifying actionn
    TYPE(fixed_formifier_visitor) :: fixed_visitor
    
    ! The object that encapsulates the freeformifying action.
    TYPE(free_formifier_visitor) :: free_visitor
    
    ! Source form for reading from the console.
    CLASS(SourceForm), ALLOCATABLE :: console_form
    
    ! Source object connected to the input unit.
    TYPE(SourceUnit), POINTER :: source_unit
    
    ! Pointer to source_unit.
    CLASS(Source), POINTER :: p_source
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Prepare the list of file names
    
    ! Convert the list of file specifications to a list of file names.  In 
    ! many cases this will be a simple copy.
    CALL ExpandSpecsToNames(host, file_list, filenames, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_err) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Build the statement tree.
    
    ALLOCATE(combined)
    ALLOCATE(combined%prog_units(0))
    
    DO if = 1, SIZE(filenames)
      CALL DoBuildTreePass(  &
          host,  &
          filenames(if)%item,  &
          current,  &
          err_list )
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_err) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_err) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END DO
    
    ! If enabled, read statements from the console.
    IF (console_read /= '') THEN
      CALL ConnectSourceUnit(INPUT_UNIT, source_unit)
      p_source => source_unit
      SELECT CASE (console_read)
      CASE ('FIXED')
        ALLOCATE(console_form, SOURCE=FixedFormSource(p_source))
      CASE ('FREE')
        ALLOCATE(console_form, SOURCE=FreeFormSource(p_source))
      END SELECT
      
      CALL DoBuildTreePass(  &
          host,  &
          console_form,  &
          current,  &
          err_list)
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_err) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_err) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (free) THEN
      ! Set the logical unit for output.
      free_visitor%unit = out_unit
      free_visitor%bad_lines = [INTEGER ::]
      
      ! Traverse the tree.
      CALL combined%Query(free_visitor)
      
      ! Until we know better.
      stop_code = 0
    ELSE
      ! Set the logical unit for output.
      fixed_visitor%unit = out_unit
      
      ! Traverse the tree.
      CALL combined%Query(fixed_visitor)
      
      IF (fixed_visitor%bad) THEN
        WRITE (err_unit, "('Error: One or more statements were too long to &
            &be represented as fixed form source.')")
        stop_code = 2
      ELSE
        stop_code = 0
      END IF
    END IF
    
    !---------------------------------------------------------------------------
    
    DEALLOCATE(combined)
    
    CALL ReleaseSources
    
  END SUBROUTINE Execute
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Fixed form stuff.
  
  
  !*****************************************************************************
  !!
  !> Implementation of fixed_formifier_visitor%ExecuteNode.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     st                The statement tree node.
  !!
  !! This is a no-op.
  
  SUBROUTINE ffv_ExecuteNode(visitor, st)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(fixed_formifier_visitor), INTENT(INOUT) :: visitor
    CLASS(StNode), INTENT(IN), TARGET :: st
    
    !***************************************************************************
    
    ! Nothing to do.
    
  END SUBROUTINE ffv_ExecuteNode
  
  
  !*****************************************************************************
  !!
  !> Implementation of fixed_formifier_visitor%ExecuteStmt.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     parent            The statement tree node that owns 
  !! the statement.
  !!
  !! @param[in]     stmt              The statement data.
  !!
  !! The approach we take is to render the statement to a linear character 
  !! buffer, the test whether the concatenation of that buffer to the 
  !! internal block buffer of the visitor object would trigger a need to 
  !! start a new initial line.  If so, write the current block buffer to the 
  !! output and reset its content.  Either way, then append the linear 
  !! buffer to the block buffer.  Finally test whether the appended 
  !! statement is an end statement and if so, write and reset the block buffer.
  !!
  !! This last test both avoids the potential for a continued end statement 
  !! (unless the user has gone crazy with the name of the program unit and 
  !! has included that program unit name in the end statement) and also 
  !! ensures that tree traversal ends with an empty buffer - we only 
  !! traverse complete program units and complete program units always end 
  !! with an end statement for a program unit!
  
  SUBROUTINE ffv_ExecuteStmt(visitor, parent, stmt)
    
    USE Statements
    USE StatementData
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(fixed_formifier_visitor), INTENT(INOUT) :: visitor
    CLASS(StNode), INTENT(IN), TARGET :: parent
    TYPE(StData), INTENT(IN) :: stmt
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! The statement rendered as a single line of fixed form text.
    CHARACTER(:), ALLOCATABLE :: source_text
    
    !***************************************************************************
    
    ! Get the fixed form representation of the source as a single line.
    CALL get_fixed_form_text(stmt%tlist, source_text)
    
    ! See if we need to flush the current buffer to the output before we add 
    ! this statement.
    IF (need_to_flush()) CALL flush_buffer(visitor)
    
    ! Save the statement label, if there is one.  (If there is one the 
    ! buffer will also have been flushed above.)
    IF (ALLOCATED(stmt%statement_label)) THEN
      visitor%label = .ValueOf. stmt%statement_label
    END IF
    
    ! Add this statement to the buffer.
    CALL append_to_buffer(visitor, source_text)
    
    ! If the statement added was an end statement for a program unit, then 
    ! flush the buffer.  This both stops us trying to continue the end 
    ! statement of a program unit (3.3.3.5) and putting additional statements 
    ! after the end statement of a program unit (prevented by the definition 
    ! of a program unit being "a sequence of one or more lines" (3.3.1p1)).
    IF (StMatches(stmt%ist, isgProgramUnitEnd)) THEN
      IF ( is_program_unit(parent)  &
          .OR. isg_program_unit_end_never_followed ) THEN
        CALL flush_buffer(visitor)
      END IF
    END IF
    
  CONTAINS
    
    !***************************************************************************
    !
    ! Utility test that determines whether the current contents of the 
    ! buffer need to be flushed.
    
    FUNCTION need_to_flush()
      
      ! Function result.
      LOGICAL :: need_to_flush
      
      !*************************************************************************
      
      ! If we have recently flushed, no need to flush now.
      need_to_flush = visitor%pos /= 0
      IF (.NOT. need_to_flush) RETURN
      
      ! If we have a statement label then we need to flush.
      need_to_flush = ALLOCATED(stmt%statement_label)
      IF (need_to_flush) RETURN
      
      ! If the current statement would be too long for the buffer, then we 
      ! need to flush.  The `+ 1` is for the statement separator character.
      need_to_flush = visitor%pos + LEN(source_text) + 1 > LEN(visitor%buffer)
      IF (need_to_flush) RETURN
      
      ! If the separator character would be the first character in a 
      ! continuation line, then we need to flush (the first character on a 
      ! line cannot be a ';' - 3.3.3.4p2).
      need_to_flush = MOD(visitor%pos,statement_field_length) == 0
      IF (need_to_flush) RETURN
      
      ! If the current statement is the end statement for a program unit and 
      ! it looks like the current statement will need to be continued, then 
      ! we need to flush (the end statement of a program unit cannot be 
      ! continued).
      IF (StMatches(stmt%ist, isgProgramUnitEnd)) THEN
        IF ( is_program_unit(parent)  &
            .OR. isg_program_unit_end_never_continued ) THEN
          ! Is the current line different from the future line?
          need_to_flush = spans_two_lines(visitor, source_text)
          IF (need_to_flush) RETURN
        END IF
      END IF
      
      ! If the initial line of the current statement "appears" like a program 
      ! unit end statement we flush.  Technically this is pessimistic - 
      ! we just need to flush enough of the source such that the initial 
      ! line of the statement doesn't look like END - say two characters.
      !
      ! But the saving is small and the complexity hassle would be great.
      CALL test_appears_like_end(visitor, parent, source_text, need_to_flush)
      IF (need_to_flush) RETURN
      
    END FUNCTION need_to_flush
    
  END SUBROUTINE ffv_ExecuteStmt
  
  
  !*****************************************************************************
  !!
  !> Test whether the initial line of some source "appears to be a 
  !! program unit END statement.
  !!
  !! @param[in]     visitor           The object carrying out the operation.
  !!
  !! @param[in]     st                The current statement tree node.
  !!
  !! @param[in]     source_text       The source.  Only the initial line of 
  !! this is checked.
  !!
  !! @param[out]    end_like          Flag that indicates whether the 
  !! initial line of the source appears end like.
  !!
  !! The consider_context_for_appears_like_end module constant affects 
  !! what we consider end like.
  
  SUBROUTINE test_appears_like_end(visitor, st, source_text, end_like)
    
    USE CharacterTypes
    USE CharUtils
    USE ProgUnitStNodes
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(fixed_formifier_visitor), INTENT(IN) :: visitor
    CLASS(StNode), INTENT(IN) :: st
    CHARACTER(*), INTENT(IN) :: source_text
    LOGICAL, INTENT(OUT) :: end_like
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: char_count     ! Number of characters in initial line.
    
    ! Initial line of the statement.
    CHARACTER(:), ALLOCATABLE :: initial_text
    
    ! The statement tree node for the program unit enclosing the current 
    ! statement.
    CLASS(ProgUnitStNode), POINTER :: prog_unit
    
    ! Text for the end program unit statement for the current program unit.
    CHARACTER(:), ALLOCATABLE :: end_prog_unit_text
    
    ! Program unit name.
    CHARACTER(:), ALLOCATABLE :: prog_unit_name
    
    INTEGER :: i              ! Character index.
    
    !***************************************************************************
    
    ! If the statement doesn't span two lines, then it can't "appear" like 
    ! a program unit end statement unless it IS a program unit end statement, 
    ! in which case we've already eliminated it.
    end_like = spans_two_lines(visitor, source_text)
    IF (.NOT. end_like) RETURN
    
    ! If there are less than three characters of the statement in the initial 
    ! line, then we don't have something that appears like an end statement.
    char_count = initial_char_count(visitor, source_text)
    end_like = .NOT. (char_count < 3)
    IF (.NOT. end_like) RETURN
    
    ! Extract the initial line.  This will have three or more characters.
    initial_text = source_text(:char_count)
    
    ! In case we change our mind about the case being spat out by 
    ! get_fixed_form_text.
    CALL UpperCase(initial_text)
    
    ! First three characters have to be END.
    end_like = .NOT. (initial_text(:3) /= 'END')
    IF (.NOT. end_like) RETURN
    
    ! END is very much like END.
    end_like = LEN(initial_text) == 3
    IF (end_like) RETURN
    
    ! Not many cases should make it to here.
    
    ! Any characters other than a letter or a digit or underbar means 
    ! it doesn't appear to be an end statement.
    DO i = 1, LEN(initial_text)
      end_like = IsNameChar(initial_text(i:i))
      IF (.NOT. end_like) RETURN
    END DO
    
    IF (consider_context_for_appears_like_end == 0) THEN
      end_like = (INDEX(initial_text, 'ENDBLOCKDATA') == 1)  &
          .OR. (INDEX(initial_text, 'ENDFUNCTION') == 1)  &
          .OR. (INDEX(initial_text, 'ENDMODULE') == 1)  &
          .OR. (INDEX(initial_text, 'ENDPROGRAM') == 1)  &
          .OR. (INDEX(initial_text, 'ENDSUBMODULE') == 1)  &
          .OR. (INDEX(initial_text, 'ENDSUBROUTINE') == 1)
      RETURN
    END IF
    
    ! Considering context from here on.
    CALL get_program_unit_node(st, prog_unit)
    SELECT CASE (prog_unit%GetPart())
    CASE (iptMainProgram)
      end_prog_unit_text = 'ENDPROGRAM'
    CASE (iptBlockData)
      end_prog_unit_text = 'ENDBLOCKDATA'
    CASE (iptFunctionSubprogram)
      end_prog_unit_text = 'ENDFUNCTION'
    CASE (iptSubroutineSubprogram)
      end_prog_unit_text = 'ENDSUBROUTINE'
    CASE (iptModule)
      end_prog_unit_text = 'ENDMODULE'
    CASE (iptSubModule)
      end_prog_unit_text = 'ENDSUBMODULE'
    CASE DEFAULT
      STOP 'In FixedFormForEvenImplementation::test_appears_like_end, &
          &encountered an unknown program unit type.'
    END SELECT
    end_like = INDEX(initial_text, end_prog_unit_text) == 1
    IF (.NOT. end_like) RETURN    ! Lead in bit is wrong.
    
    ! If consider_xxx is 1, then we don't about the name bit.
    IF (consider_context_for_appears_like_end == 1) RETURN
    
    ! Here if we've asked for name checking too.
    
    ! Is there a name on the possible end statement?
    end_like = LEN(end_prog_unit_text) == LEN(initial_text)
    IF (end_like) RETURN          ! No.
    
    ! Can we get a program unit name?
    CALL prog_unit%GetName(prog_unit_name)
    IF (.NOT. ALLOCATED(prog_unit_name)) THEN
      STOP 'In FixedFormForEvenImplementation::test_appears_like_end, &
          &the program unit name could not be determined.'
    END IF
    
    ! If the name off the opening statement is zero length (there was no 
    ! name) then we can't be looking for a name here.
    end_like = .NOT. (LEN(prog_unit_name) == 0) 
    IF (.NOT. end_like) RETURN
    
    ! Do we have a match with the name appended?
    end_like = (end_prog_unit_text // prog_unit_name) == initial_text
    
  END SUBROUTINE test_appears_like_end
  
  
  !*****************************************************************************
  !!
  !> Get the program unit ancestor of the current statement tree node.
  !!
  !! @param[in]     st                The current statement tree node.
  !!
  !! @param[out]    prog_unit         Pointer to the parent ancestor of 
  !! @a st that is a program unit statement tree node.
  !!
  !! This involves searching back up the tree until we find a node that is a 
  !! child of a BaseStNode, then calling GetPart on that child node.
  
  SUBROUTINE get_program_unit_node(st, prog_unit)
    
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StNode), INTENT(IN) :: st
    CLASS(ProgUnitStNode), INTENT(OUT), POINTER :: prog_unit
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Pointer for walking up the tree.
    CLASS(StNode), POINTER :: tmp
    
    !***************************************************************************
    
    tmp => st%parent
    DO WHILE (ASSOCIATED(tmp))
      IF (is_program_unit(tmp)) THEN
        ! Found it.  Cast the result to the desired type.
        SELECT TYPE (tmp)
        CLASS IS (ProgUnitStNode)
          prog_unit => tmp
          RETURN
        CLASS DEFAULT
          STOP 'In FixedFormForEvenImplementation::get_program_unit_node, &
            &the program unit node wasn''t a program unit node.'
        END SELECT
      END IF
      tmp => tmp%parent
    END DO
    
    ! Shouldn't happen, unless the tree is borked.
    STOP 'In FixedFormForEvenImplementation::get_program_unit_node, &
        &couldn''t find the program unit node.'
    
  END SUBROUTINE get_program_unit_node
  
  
  !*****************************************************************************
  !!
  !> Test whether a statement spans two lines, assuming that this is not 
  !! the first statement on the line.
  !!
  !! @param[in]     visitor           The object carrying out the operation.
  !!
  !! @param[in]     source_text       The text of the statement.
  
  FUNCTION spans_two_lines(visitor, source_text)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(fixed_formifier_visitor), INTENT(IN) :: visitor
    CHARACTER(*), INTENT(IN) :: source_text
    
    ! Function result.
    LOGICAL :: spans_two_lines
    
    !***************************************************************************
    
    ! We add one to visitor%pos to account for the semicolon.
    spans_two_lines = (visitor%pos + 1) / statement_field_length  &
        /= (visitor%pos + LEN(source_text) + 1) / statement_field_length
    
  END FUNCTION spans_two_lines
  
  
  !*****************************************************************************
  !!
  !> Get the number of characters of a statement that are on the initial line, 
  !! assuming the statement spans multiple lines.
  !!
  !! @param[in]     visitor           The object carrying out the operation.
  !!
  !! @param[in]     source_text       The text of the statement.
  
  FUNCTION initial_char_count(visitor, source_text)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(fixed_formifier_visitor), INTENT(IN) :: visitor
    CHARACTER(*), INTENT(IN) :: source_text
    
    ! Function result.
    INTEGER :: initial_char_count
    
    !***************************************************************************
    
    ! The + 1 is for the semicolon.
    initial_char_count = MIN( LEN(source_text),  &
        statement_field_length - MOD(visitor%pos + 1, statement_field_length) )
    
  END FUNCTION initial_char_count
  
  
  !*****************************************************************************
  !!
  !> Write the current buffer to the output unit.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE flush_buffer(visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(fixed_formifier_visitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: pos            ! Position of last character written.
    
    ! Format specification for the initial line when there is a label.
    !
    ! Big labels (six digits) will use the comment indicator field.
    CHARACTER(*), PARAMETER :: fmt_initial_line_with_label = "(I5,1X,A)"
    
    ! Format specification for the initial line when there is no label.
    CHARACTER(*), PARAMETER :: fmt_initial_line_no_label = "(5X,1X,A)"
    
    ! Format specification - skip over the label field.
    CHARACTER(*), PARAMETER :: fmt_continued_line = "(5X,A,A)"
    
    ! Valid characters for continuation.  No blank or zero.
    CHARACTER(*), PARAMETER :: continue_chars  &
        = '123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ=+-*/\()[]{},.:;!"%&~<>?''`^|$#@'
    
    REAL :: r                 ! A random number.
    INTEGER :: ic             ! continue char index.
    
    !***************************************************************************
    
    ! Nothing in the buffer means nothing to do.
    IF (visitor%pos == 0) RETURN
    
    ! Initial line.
    pos = statement_field_length
    IF (visitor%label /= 0) THEN
      WRITE (visitor%unit, fmt_initial_line_with_label) visitor%label,  &
          visitor%buffer(:pos)
    ELSE
      WRITE (visitor%unit, fmt_initial_line_no_label)  &
          visitor%buffer(:pos)
    END IF
    
    DO WHILE (pos < visitor%pos)
      CALL RANDOM_NUMBER(r)
      ic = FLOOR(r * LEN(continue_chars)) + 1
      WRITE (visitor%unit, fmt_continued_line) continue_chars(ic:ic),  &
          visitor%buffer(pos+1:pos+statement_field_length)
      pos = pos + statement_field_length
    END DO
    
    ! Reset buffer.
    visitor%buffer = ''
    visitor%pos = 0
    visitor%label = 0
    
  END SUBROUTINE flush_buffer
  
  
  !*****************************************************************************
  !!
  !> Append text to the buffer.
  !! 
  !! @param[in,out] visitor           The object carrying out the operation.
  !!
  !! @param[in]     source_text       The text to add to the buffer.
  !!
  !! A semicolon is added before the given text if the buffer is not empty.
  !!
  !! The buffer position is updated.  The text to be added had better fit.
  
  SUBROUTINE append_to_buffer(visitor, source_text)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(fixed_formifier_visitor), INTENT(INOUT) :: visitor
    CHARACTER(*), INTENT(IN) :: source_text
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: last           ! last character position.
    
    !***************************************************************************
    
    ! If we have stuff in the buffer then we need a separating semicolon.
    IF (visitor%pos /= 0) THEN
      visitor%pos = visitor%pos + 1
      visitor%buffer(visitor%pos:visitor%pos) = ';'
    END IF
    
    ! Append the text.
    !
    ! This sanity check should only trigger if the length of a single 
    ! statement exceeds that permitted for fixed form source.  This is 
    ! possible if warnings from the scanner are not regarded or if the 
    ! incoming source is free form.
    last = visitor%pos + LEN(source_text)
    IF (last > LEN(visitor%buffer)) THEN
      visitor%bad = .TRUE.
      last = LEN(visitor%buffer)
    END IF
    
    visitor%buffer(visitor%pos+1:last) = source_text
    visitor%pos = last
    
  END SUBROUTINE append_to_buffer
  
  
  !*****************************************************************************
  !!
  !> Get the statement as fixed form text.
  !!
  !! @param[in]     tlist             The sequence of tokens for the statement.
  !!
  !! @param[out]    text              The fixed form text that corresponds to 
  !! the sequence of tokens.
  !!
  !! We emit no spaces between tokens and capitalise anything that isn't 
  !! a character literal.
  
  SUBROUTINE get_fixed_form_text(tlist, text)
    
    USE CharUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CHARACTER(:), INTENT(OUT), ALLOCATABLE :: text
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Token index.
    
    ! Raw token text.
    CHARACTER(:), ALLOCATABLE :: raw_text
    
    !***************************************************************************
    
    text = ''
    DO i = 1, SIZE(tlist)
      raw_text = QueryRaw(tlist(i))
      IF ( (QueryType(tlist(i)) /= ittCharacterLiteral)  &
          .AND. (QueryType(tlist(i)) /= ittFormatSpecification) ) THEN
        CALL UpperCase(raw_text)
      END IF
      text = text // raw_text
    END DO
    
  END SUBROUTINE get_fixed_form_text
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Free form stuff.
  
  
  !*****************************************************************************
  !!
  !> Implementation of free_formifier_visitor%ExecuteNode.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     st                The statement tree node.
  
  SUBROUTINE freev_ExecuteNode(visitor, st)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(free_formifier_visitor), INTENT(INOUT) :: visitor
    CLASS(StNode), INTENT(IN), TARGET :: st
    
    !***************************************************************************
    
    ! Avoid not used.
    IF (LEN(visitor%buffer) /= 0) CONTINUE
    IF (ASSOCIATED(st%parent)) CONTINUE
    
    ! Nothing to do.
    
  END SUBROUTINE freev_ExecuteNode
  
  
  !*****************************************************************************
  !!
  !> Implementation of free_formifier_visitor%ExecuteStmt.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     parent            The statement tree node that owns 
  !! the statement.
  !!
  !! @param[in]     stmt              The statement data.
  !!
  !! We get the text of each token (or label) and consider whether a space 
  !! is required before the previous token.  Then we consider whether the token 
  !! text (possibily including leading space) will fit in the 
  !! buffer.  If not, we write the buffer and reset its contents.  Then we 
  !! consider whether the token will be split.  Then we reconsider our 
  !! approach and stop writing this comment.
  
  SUBROUTINE freev_ExecuteStmt(visitor, parent, stmt)
    
    USE CharUtils
    USE Statements
    USE StatementData
    USE SyntaxParts
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(free_formifier_visitor), INTENT(INOUT) :: visitor
    CLASS(StNode), INTENT(IN), TARGET :: parent
    TYPE(StData), INTENT(IN) :: stmt
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: pass           ! Pass through the file.
    INTEGER :: column         ! Token specific position (column) in the buffer.
    INTEGER :: line           ! Token specific position (line) in the buffer.
    INTEGER :: it             ! Token index.
    LOGICAL :: last           ! Flag for the last token in the statement.
    
    ! Buffer for formatting labels.  R312 limits this to five characters.
    CHARACTER(5) :: label
    
    ! Text of the current token.
    CHARACTER(:), ALLOCATABLE :: token_text
    
    !***************************************************************************
    
    ! The first pass through we may not be dealing with an empty buffer.  
    ! The second pass through we definitely are (or want to be) because the 
    ! first pass failed.
    DO pass = 1, 2
      IF (pass == 2) CALL free_flush_buffer(visitor)
      column = visitor%column
      line = visitor%line
      
      ! Do we need a separating ';'?
      IF (.NOT. buffer_empty(visitor)) THEN
        ! Never here on pass two.
        IF (visitor%column == 0) THEN
          ! We are at the start of the line anyway.  Because of the last 
          ! token indicator when adding text to the buffer, the previous 
          ! line will not end with an ampersand - so just start a new 
          ! initial line.
          CYCLE
        ELSE IF (visitor%column >= LEN(visitor%buffer) - 2) THEN
          ! There's no point putting a semicolon and a continuation ampersand 
          ! at the very end of the previous line or an ampersand at the end 
          ! and a semicolon at the start of the next line - just start a 
          ! new initial line.
          CYCLE
        END IF
        CALL add_text(visitor, column, line, ';', .FALSE.)
        ! This should never trigger because of the elseif condition immediately 
        ! above, but it is here for completeness.
        IF (column > LEN(visitor%buffer)) CYCLE
      END IF
      
      ! Do we have a label?
      IF (ALLOCATED(stmt%statement_label)) THEN
        WRITE (label, "(I0)") .ValueOf. stmt%statement_label
        ! Labels cannot be on their own, they always need to have a 
        ! statement following them.
        CALL add_text(visitor, column, line, TRIM(label), .FALSE.)
        ! As of F08/0075, statements always start with a name or keyword, 
        ! hence we always need a space after a label.
        CALL add_text(visitor, column, line, ' ', .FALSE.)
        IF (column > LEN(visitor%buffer)) CYCLE
      END IF
      
      ! Run through the tokens.
      DO it = 1, SIZE(stmt%tlist)
        last = it == SIZE(stmt%tlist)
        
        ! Get the text for the token in token_text.
        token_text = QueryRaw(stmt%tlist(it))
        IF ( (stmt%tlist(it) /= ittCharacterLiteral)  &
            .AND. (stmt%tlist(it) /= ittFormatSpecification) ) THEN
          CALL LowerCase(token_text)
        END IF
        
        ! Put that text to the buffer.
        CALL add_text(visitor, column, line, token_text, last)
        IF (column > LEN(visitor%buffer)) CYCLE
        
        ! Will we need a separating blank before the next token?
        IF (.NOT. last) THEN
          IF (blank_required(stmt%tlist(it), stmt%tlist(it+1))) THEN
            CALL add_text(visitor, column, line, ' ', last)
            IF (column > LEN(visitor%buffer)) CYCLE
          END IF
        END IF
      END DO
      ! We processed all the tokens and no-one complained.  We are done.  
      ! Because the loop iteration control stuff won't be executed, this means 
      ! that pass cannot be three if things have worked.
      EXIT
    END DO
    
    ! Was a single statement too large to fit into a single block (which 
    ! raises the question as to how the statement got parsed in in the first 
    ! place, but nevermind).
    IF (pass == 3) THEN
      ! It was too large!
      ! Flush what we have and note the problem.
      visitor%line = SIZE(visitor%buffer)
      visitor%column = LEN(visitor%buffer)
      CALL free_flush_buffer(visitor)
      visitor%bad_lines = [visitor%bad_lines, 1] ! stmt%list(1)%GetLocation
    ELSE
      ! The statement fitted nicely thank you.  Update the line and column 
      ! components inside the visitor object to accept what we've just 
      ! written.
      visitor%line = line
      visitor%column = column
      
      ! The definition of a program unit is a sequence of lines, so if this 
      ! statement ends a program unit, then we need to flush so that 
      ! we don't try and write the next program unit on the same line.
      !
      ! A program unit node (as considered by is_program_unit) won't 
      ! present a statement that could end a program unit other than the 
      ! one that does actually end the program unit.  Where 
      ! program-unit-ending-like statements are used to end internal 
      ! or module subprograms or interface bodies, those statements will 
      ! be presented by statement tree nodes that are not program units 
      ! (they might be ProgUnitStNodes, but they won't be children of the 
      ! base node).
      IF (StMatches(stmt%ist, isgProgramUnitEnd)) THEN
        IF (is_program_unit(parent)) CALL free_flush_buffer(visitor)
      END IF
    END IF
    
  END SUBROUTINE freev_ExecuteStmt
  
  
  !*****************************************************************************
  !!
  !> Write the current contents of the buffer, up to the character indicated 
  !! by the line and column components, to the output.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE free_flush_buffer(visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(free_formifier_visitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Line index.
    
    !***************************************************************************
    
    ! Write out the complete lines.
    DO i = 1, visitor%line - 1
      WRITE (visitor%unit, "(A)") visitor%buffer(i)
    END DO
    
    ! Write out the last incomplete line.
    IF (visitor%column /= 0) THEN
      ! ifort bug - array character components of extended type.
      ASSOCIATE(buf => visitor%buffer(i))
        WRITE (visitor%unit, "(A)") buf(1:visitor%column)
      END ASSOCIATE
    END IF
    
    ! Clear and reset buffer indices.
    visitor%buffer = ''
    visitor%line = 1
    visitor%column = 0
    
  END SUBROUTINE free_flush_buffer
  
  
  !*****************************************************************************
  !!
  !> Add some text to the buffer.
  !!
  !! @param[in]     visitor           The object carrying out the operation.
  !!
  !! @param[in,out] column            The last column written to in the 
  !! buffer for the current statement.  This is different to visitor%column, 
  !! which is the last column written to in the buffer for the previous 
  !! statement.  This is updated by this call.  After the call, if this 
  !! is equal to LEN(buffer), then the text exactly filled the buffer.  If 
  !! this is greater than LEN(buffer), then the text had to be truncated 
  !! in order to fill the buffer.
  !!
  !! @param[in]     line              The next line to be written to for 
  !! the current statement.
  !!
  !! @param[in]     text              The text to be placed in the buffer.
  !!
  !! @param[in]     last              Flag to indicate that this is the text 
  !! for the last token in the statement.  For text for tokens other than the 
  !! last, the last character of the line is reserved for the statement 
  !! continuation ampersand.
  
  SUBROUTINE add_text(visitor, column, line, text, last)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(free_formifier_visitor), INTENT(INOUT) :: visitor
    INTEGER, INTENT(INOUT) :: column
    INTEGER, INTENT(INOUT) :: line
    CHARACTER(*), INTENT(IN) :: text
    LOGICAL, INTENT(IN) :: last
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: pos            ! Last position in text added to buffer.
    INTEGER :: num            ! Length of the current stretch of text.
    
    !***************************************************************************
    
    ! If we are already full, then just return without doing anything.  This 
    ! allows calling code to call us multiple times for a related stretch of 
    ! text that has to go in the same block of code, and then test for 
    ! fullness at the end.
    IF (column >= LEN(visitor%buffer)) RETURN
    
    pos = 0
    
    DO WHILE (pos < LEN(text))
      ! The merger reserves the last column for the statement continuation 
      ! ampersand.  We suppress the allowance for that ampersand on the 
      ! last line of the buffer.
      num = MIN( LEN(text) - pos, LEN(visitor%buffer) - column  &
          - MERGE(0, 1, last .AND. line < SIZE(visitor%buffer)) )
      IF ( last  &
          .AND. (num + pos < LEN(text)) &
          .AND. (line < SIZE(visitor%buffer)) ) THEN
        ! We have to split the token and hence we need an ampersand at the end 
        ! of the line.  We didn't allow for that in the calculation of num 
        ! above.  This might make num zero, but that is ok.
        !
        ! If this is the last line in the buffer, then we do not want to 
        ! append a continuation char - we simply truncate the last token.
        num = num - 1
      END IF
      ! Transfer the num characters to the buffer.
      ! ifort bug - array character components of extended type.
      ASSOCIATE(buf => visitor%buffer(line))
        buf(column+1:column+num) = text(pos+1:pos+num)
      END ASSOCIATE
      column = column + num
      ! Do we need a coninuation or initial splitting ampersand?
      IF (line < SIZE(visitor%buffer)) THEN
        IF (num + pos < LEN(text)) THEN
          ! Token splitting.
          ! ifort bug - array character components of extended type.
          ASSOCIATE(buf => visitor%buffer(line))
            buf(column+1:column+1) = '&'
          END ASSOCIATE
          column = 1
          line = line + 1
          ! Final splitting ampersand.
          ! ifort bug - array character components of extended type.
          ASSOCIATE(buf => visitor%buffer(line))
            buf(column:column) = '&'
          END ASSOCIATE
        ELSE IF ( .NOT. last  &
            .AND. (column == LEN(visitor%buffer) - 1) ) THEN
          ! Continuation ampersand.
          column = LEN(visitor%buffer)
          ! ifort bug - array character components of extended type.
          ASSOCIATE(buf => visitor%buffer(line))
            buf(column:column) = '&'
          END ASSOCIATE
          column = 0
          line = line + 1
        END IF
      ELSE IF (num + pos < LEN(text)) THEN
        ! We are on the last line of the buffer and we couldn't fit the 
        ! text.  Indicate that by setting column to an out of range value 
        ! and quitting.
        column = LEN(visitor%buffer) + 1
        RETURN
      END IF
      pos = pos + num
    END DO
    
  END SUBROUTINE add_text
  
  
  !*****************************************************************************
  !!
  !> Is the buffer empty?
  !!
  !! @param[in]     visitor           The object carrying out the operation.
  !!
  !! @returns true if it is, false otherwise.
  
  FUNCTION buffer_empty(visitor)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(free_formifier_visitor), INTENT(IN) :: visitor
    
    ! Function result.
    LOGICAL :: buffer_empty
    
    !***************************************************************************
    
    buffer_empty = visitor%line == 1 .AND. visitor%column == 0
    
  END FUNCTION buffer_empty
  
  
  !*****************************************************************************
  !!
  !> Determine whether a blank is required to separate two tokens in free 
  !! form source.
  !!
  !! @param[in]     left              The token on the left.
  !!
  !! @param[in]     right             The token on the right.
  !!
  !! @returns True if the criteria in 3.3.2.2p2 is met - i.e. "A blank shall 
  !! be used to separate names, constants or labels from adjacent keywords, 
  !! names, constants or labels."
  !!
  !! We also test for the adjacent keyword case where blanks are optional - 
  !! returning false when we can unambiguously get away without a blank.
  !!
  !! Not written to accomodate token type modification by the parse pass.
  
  FUNCTION blank_required(left, right)
    
    USE CompilerKinds
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: left
    TYPE(Token), INTENT(IN) :: right
    
    ! Function result.
    LOGICAL :: blank_required
    
    !***************************************************************************
    
    IF ((left == ittName) .AND. (right == ittName)) THEN
      ! This is the tricky one (unless we want to bring the parse pass in 
      ! as well as the build tree pass, which we don't).  Consider a 
      ! /case-construct/ with a construct name of "case".  The terminating 
      ! /end-select-stmt/ could be END SELECT CASE, so if we were to just 
      ! blindly go off and stick all occurrences of SELECT and CASE together, 
      ! as 3.3.2.2 might suggest, we'd be in trouble.  Examining additional 
      ! tokens prior to @a left might help, but that's starting to get 
      ! complicated.  Just deal with the unambiguous cases.
      !
      ! Ambiguous case are BLOCK DATA, ELSE IF, ELSE WHERE, SELECT TYPE and
      ! SELECT CASE.
      IF (left == scck_'END') THEN
        ! Not efficient, but it works.
        blank_required = .NOT. ( (right == scck_'ASSOCIATE')  &
            .OR. (right == scck_'BLOCK')  &
            .OR. (right == scck_'BLOCKDATA')  &
            .OR. (right == scck_'CRITICAL')  &
            .OR. (right == scck_'DO')  &
            .OR. (right == scck_'ENUM')  &
            .OR. (right == scck_'FILE')  &
            .OR. (right == scck_'FORALL')  &
            .OR. (right == scck_'FUNCTION')  &
            .OR. (right == scck_'IF')  &
            .OR. (right == scck_'INTERFACE')  &
            .OR. (right == scck_'MODULE')  &
            .OR. (right == scck_'PROCEDURE')  &
            .OR. (right == scck_'PROGRAM')  &
            .OR. (right == scck_'SELECT')  &
            .OR. (right == scck_'SUBMODULE')  &
            .OR. (right == scck_'SUBROUTINE')  &
            .OR. (right == scck_'TYPE')  &
            .OR. (right == scck_'WHERE') )
      ELSE IF (left == scck_'DOUBLE') THEN
        blank_required = .NOT. (right == scck_'PRECISION')
      ELSE IF (left == scck_'GO') THEN
        blank_required = .NOT. (right == scck_'TO')
      ELSE IF (left == scck_'IN') THEN
        blank_required = .NOT. (right == scck_'OUT')
      ELSE
        blank_required = .TRUE.
      END IF
      RETURN
    ELSE
      ! Constant includes the /complex literal-constant/, which is not a 
      ! single token.  If there was a case where you could have a 
      ! complex literal constant up against a name or label then a separating 
      ! blank would be required even though the complex constant is delimited 
      ! by things that normally can run up against other tokens without 
      ! blanks, but I don't think the syntax permits such a case subsequent 
      ! to interp F08/0075.
      blank_required = is_name_constant_or_label(left)  &
          .AND. is_name_constant_or_label(right)
    END IF
    
  END FUNCTION blank_required
  
  
  !*****************************************************************************
  !!
  !> Test whether a token is a name, constant or label.
  !!
  !! @param[in]     the_token         The token to test.
  !!
  !! @returns true if the token is a name, constant or label.
  !!
  !! Written to accomodate token type modification by the parse pass.
  
  FUNCTION is_name_constant_or_label(the_token) RESULT(b)
    
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: the_token
    
    LOGICAL :: b
    
    !***************************************************************************
    
    b = (the_token == ittName)  &
        .OR. (the_token == ittKeyword)  &   ! redundant... but maybe one day!
        .OR. (the_token == ittIntegerLiteral)  &
        .OR. (the_token == ittRealLiteral)  &
        .OR. (the_token == ittCharacterLiteral)  &
        .OR. (the_token == ittLogicalLiteral)  &
        .OR. (the_token == ittBOZLiteral)  &
        .OR. (the_token == ittStatementLabel)   ! Never gonna happen.
    
  END FUNCTION is_name_constant_or_label
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Common helpers.
  
  
  !*****************************************************************************
  !!
  !> Test whether the current node is a program unit.
  !!
  !! @param[in]     node      The statement tree node to test.
  !!
  !! @returns True if the given node represents a program unit, false 
  !! otherwise.
  !!
  !! A program unit tree node is a child of a node of type BaseStNode.
  
  FUNCTION is_program_unit(node)
    
    USE BaseStNodes
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(StNode), INTENT(IN) :: node
    
    ! Function result.
    LOGICAL :: is_program_unit
    
    !***************************************************************************
    
    SELECT TYPE (p => node%parent)
    TYPE IS (BaseStNode)
      is_program_unit = .TRUE.
    CLASS DEFAULT
      is_program_unit = .FALSE.
    END SELECT
    
  END FUNCTION is_program_unit
  
END MODULE FixedFormForEverImplementation
